; ----------------------------------------------------------------------
; RetroForth 7.0             | This is the core of RetroForth. It has
; http://www.retroforth.asm  | almost all of the code needed to handle
;                            | a Forth system. Just provide KEY, EMIT,
;                            | and BYE.
; ----------------------------------------------------------------------
init_retroforth:
	mov esi, s0				; Set stack location
	upsh retroforth_f			; Start of bootstrap
	upsh retroforth_f.1 - retroforth_f	; Length of bootstrap
	call eval				; Interpret bootstrap
next
; ----------------------------------------------------------------------
interpret:	
.o:	call query		; ( Get a WORD )
.word:	upsh 32 		; ( Push a <SPACE> onto the stack )
	call parse		; ( Parse until we find <SPACE> )
	jnz .find		; ( Look for the end of the line )
	drop			; ( DROP two numbers )
	drop			;
	jmp .o			; ( Loop back to start if no words found )
.find:	call find		; ( See if we can find the word )
	jnc .exec		; ( Yes? Interpret it )
	call number		; ( No? Then make it a number )
	jnc .word		; ( Loop back )
	drop			; ( Drop two numbers )
	drop			;
	jmp .o			; ( And Loop back )
.exec:	upop edi		; ( This is where we actually call the )
	call edi		; ( words we find )
	jmp .word		; ( And Loop back )
next
; ----------------------------------------------------------------------
code 'mfind', mfind		;
	push ebx		;
	mov ebx, mlast		;
	jmp find.o		;
; ----------------------------------------------------------------------
code 'find', find		;
	push ebx		;
	mov ebx, flast		;
.o:	push ecx		;
	upop ecx		;
.a:	mov ebx,[ebx]		;
	or ebx,ebx		;
	jz .end			;
	cmp cl,[ebx+8]		;
	jne .a			;
.len:	push esi		; same length
	push edi		;
	push ecx		;
	mov esi,eax		;
	lea edi,[ebx+9]		;
	rep cmpsb		;
	pop ecx			;
	pop edi			;
	pop esi			;
	jne .a			;
	mov eax,[ebx+4] 	; exact match
	clc			;
	jmp .ret		;
.end:	upsh ecx		; no matches
	stc			;
.ret:	pop ecx			;
	pop ebx			;
next
; ----------------------------------------------------------------------
code '>number', number		;
	push dword [base]	;
	mov ecx,eax		; n   (keep on stack in case of failure)
	mov ebx,[esi]		; a
	dup			;
	xor eax,eax		; the number
	xor edx,edx		; temp
	mov dl,[ebx]		;
	cmp dl,45		; -     Sign prefix
	pushf			;
	jne .a			;
	inc ebx			;
	dec ecx			;
	mov dl,[ebx]		;
.a:	sub dl,36		; $%&'  Base prefix
	cmp dl,4		;
	ja .b			;
	mov dl,[edx+bases]	;
	mov byte [base],dl	;
	inc ebx			;
	dec ecx			;
.b:	mov dl,[ebx]		; digits
	inc ebx			;
	call digit		;
	jc .err			;
	loop .b			;
	jmp .c			;
.err:	add esp, 4		;
	drop			;
	stc			;
	jmp .ret		;
.c:	popf			;
	jne .d			;
	neg eax			;
.d:	add esi,8		;
	clc			;
.ret:	pop dword [base]	;
next
; ----------------------------------------------------------------------
digit:	cmp byte [base],255	;
	je .x10			;
	cmp dl,57		; 9
	jbe .a			;
	and dl,5Fh		; uppercase
	cmp dl,65		; throw out chars between '9' and 'A'
	jb .err			;
	sub dl,7		;
.a:	sub dl,48		; 0
	cmp dl, byte [base]	;
	jb .x10			;
.err:	stc			; not a digit
	ret			;
.x10:	imul eax,[base]		;
	add eax,edx		;
	clc			;
next
; ----------------------------------------------------------------------
query:				;
	mov dword ecx,[source]	;
	or ecx,ecx		;
	jnz query_mem		;
	mov dword [tp],tib	; Reset TP, TIN
	mov dword [tin],tib	;
.a:	call key		; Get a keypress
	cmp al, 13		; 
	je .cr			;
	cmp al, 10		;
	je .cr			;
	xchg edi,[tp]		; Store the char in the TIB
	stosb			;
	xchg [tp],edi		;
	drop			;
	jmp .a			; And Loop back around
.cr:	drop			; We just eat CR & LF
next
; ----------------------------------------------------------------------
query_mem:			;
	mov edi,[tin]		;
	upsh edi		; Input pointer
	dup			;
	cmp Byte [edi],10	; Skip LF
	jne .a			;
	inc edi 		; Increase our pointer
.a:	mov [tin],edi		;
	sub ecx,edi		; Remaining length
	jbe .eof		;
	mov al,10		; Line Feed
	repne scasb		;
	mov eax,edi		;
	jne .b			;
	dec eax			;
	cmp Byte [eax-1],13	; We make CR optional
	jne .b			;
	dec eax			;
.b:	upop [tp]		;
	drop			;
	next			;
.eof:	drop			;
	drop			;
	add esp,4		; Discard caller
	pop Dword [tp]		;
	pop Dword [tin]		;
	pop Dword [source]	;
	next
; ----------------------------------------------------------------------
code 'eval', eval		; This takes an address & count to eval
	push Dword [source]	; Save "source"
	push Dword [tin]	; Save ">in"
	push Dword [tp]		; Save "tp"
	add eax,[esi]		; 
	upop [source]		; New "source"
	upop [tin]		; New ">in"
	jmp interpret		; Interpret the memory area
next
; ----------------------------------------------------------------------
dovar:  dup			; This little snippit of code is what
	pop eax			; pushes the address of a variable to
	next			; the stack when a variable is called
; ----------------------------------------------------------------------
code ',', comma			; comma (,) saves a value to "here"
	mov ecx,4		; By default, it uses a dword (4 bytes)
.a:	mov edx,[h]		;
	mov [edx],eax		;
	drop			;
	add edx,ecx		;
	mov [h],edx		;
	next
code '1,', comma1		; comma1 (1,) saves 1 byte to "here"
	mov ecx,1		;
	jmp comma.a
code '2,', comma2		; comma2 (2,) saves 2 bytes to "here"
	mov ecx,2		;
	jmp comma.a
code '3,', comma3		; comma3 (3,) saves 3 bytes to "here"
	mov ecx,3		;
	jmp comma.a
; ----------------------------------------------------------------------
dodoes: dup			; Handle DOES>
	pop eax			;
	xchg eax,[esp]		;
	next
; ----------------------------------------------------------------------
dolit:	dup			; This is where we handle literals.
	mov eax,[esp]		;
	mov eax,[eax]		;
	add dword [esp],4	;
	next
; ----------------------------------------------------------------------
code 'create', create
	upsh 32			; Name
	call parse		;
	push ecx		;
	mov edi,[d]		; LFA
	mov ecx,[last]		; last
	mov edx,[ecx]		;
	mov [edi],edx		; LFA= [last]
	mov [ecx],edi		; last= LFA
	mov ecx,[h]		;
	mov [edi+4],ecx		; CFA= here
	mov [edi+8],al		; Length
	add edi,9		;
	upop ecx		;
	push esi		;
	mov esi,eax		;
	rep movsb		;
	mov [d],edi		; d= d+9+length
	pop esi			;
	pop ecx			;
	mov eax,dovar		; Code
	jmp compile		;
; ----------------------------------------------------------------------
mcode 'does>', does		;
	upsh pdoes		;
	call compile		;
	upsh dodoes		;
	jmp compile		;
; ----------------------------------------------------------------------
pdoes:	dup			;
	mov eax,[last]		; eax= header
	mov eax,[eax]		;
	mov eax,[eax+4]		; eax= CFA
	pop ebx			;
	sub ebx,eax		;
	sub ebx,5		;
	mov [eax+1],ebx		; Change "call dovar" to "call <code after does>"
	drop			;
	next
; ----------------------------------------------------------------------
code ']', rbracket		; This is the actual compiler loop
.word:	upsh 32			; Space
	call parse		; Scan for a word name or number
	jnz .find		; n=0 means EOL
	drop			; 2drop
	drop			;
	call query		;
	jmp .word		; Was a 'ret'
.find:	call mfind		;
	jnc .exec		; Macros
	call find		;
	jnc .com		; Regular words
	call number		;
	jnc .lit		; Numbers
	call type		; Word not found (go figure...)
	upsh '?'		; * When a word isn't found we run a
	call emit		; * code sequence identical to this
	upsh 10			; * Forth snippet:
	call emit		; *   type '? emit cr
	jmp .word		; * And continue compiling the word
.exec:	upop edi		; Execute words in 'macro'
	call edi		;
	jmp .word		;
.com:	call compile		; Compile a call to words in 'forth'
	jmp .word		;
.lit:	upsh dolit		; Compile literals
	call compile		;
	call comma		;
	jmp .word		;
; ----------------------------------------------------------------------
code 'compile', compile		; This routine compiles in CALL's to
	sub eax,[h]		; words. Address is passed in EAX
	sub eax,5		;
	upsh 0xE8		; CALL opcode
	call comma1		; 0xE8
	call comma		; ADDRESS
next
; ----------------------------------------------------------------------
mcode '[', lbracket		; Switch back to the interpreter
	add esp,4		;
next
; ----------------------------------------------------------------------
mcode ';;', ssemi		; Exit a word (; will call this!)
	mov edx,[h]		;
	sub edx,5		;
	cmp byte [edx],0xE8	; Was the last thing compiled a CALL?
	jnz .a			; No, skip the next line
	inc byte [edx]		; Yes, change to JMP
	next			; and exit
.a:	mov byte [edx+5],0xC3	; If not a CALL, compile in a RET
	inc dword [h]		;
next
; ----------------------------------------------------------------------
mcode ';', semi			; Compile in an exit to the current word
	call ssemi		; And go back to the interpreter
	jmp lbracket		;
; ----------------------------------------------------------------------
code ':', colon			; Ok, this is the entry to the compiler
	call create		; * Create a new word
	sub dword [h],5		; * Undo 'call dovar'
	jmp rbracket		; * And jump to the real compiler
; ----------------------------------------------------------------------
mcode 'literal', literal	; Compile in a literal
	upsh dolit		;
	call compile		;
	call comma		;
next
; ----------------------------------------------------------------------
code 'forth', dict_forth	; Switch to the 'forth' vocabulary
	mov dword [last], flast	;
next
; ----------------------------------------------------------------------
code 'macro', dict_macro	; Switch to the 'macro' vocabulary
	mov dword [last], mlast	;
next
; ----------------------------------------------------------------------
code '.', print			;
	push edx		;
	push edi		;
	mov edi,esi		; edi = buffer (in stack space)
	sub edi,4		;
	or eax,eax		; Negative?
	jns .a			;
	neg eax			;
	upsh '-'		;
	call emit		;
.a:	xor edx,edx		;
	div dword [base]	;
	add dl,'0'		;
	cmp dl,'9'		;
	jbe .b			;
	add dl,7+32		;
.b:	dec edi			;
	mov [edi],dl		;
	or eax,eax		;
	jnz .a			;
	mov eax,edi		; Print
	upsh esi		;
	sub eax,edi		; # of digits
	call type		;
	pop edi			;
	pop edx			;
	upsh ' '		; And add one space after
	call emit		; the number
next
; ----------------------------------------------------------------------
code 'type',type		;
	push ebx		;
	push ecx		;
	upop ecx		;
	upop ebx		;
	or ecx,ecx		; Is the count zero?
	jz .done		; If so, we can exit
.a:	push ecx		; store the count
	dup			; Duplicate it
	xor eax,eax		; Filter out junk bits
	mov al,[ebx]		; Obtain the byte to show
	inc ebx 		; increment to the next byte
	call emit		; Show the character
	pop ecx 		; Restore the count
	loop .a			; Loop, decrementing the count
.done:	pop ecx			;
	pop ebx			;
next
; ----------------------------------------------------------------------
code 'parse', parse		;
	mov edi,[tin]		; Pointer into TIB
	mov ecx,[tp]		;
	sub ecx,edi		;
	inc ecx			;
	repe scasb		;
	dec edi			;
	inc ecx			;
	dup			;
	mov [esi],edi		; a
	repne scasb		;
	mov eax,edi		;
	dec eax			;
	sub eax,[esi]		; n
	mov [tin],edi		;
next
; ----------------------------------------------------------------------
code 'words', words		;
	push ebx		;
	mov ebx, [last]		;
.a:	mov ebx,[ebx]		;
	or ebx,ebx		;
	jz .end			;
	dup			;
	lea eax,[ebx+8]		;
	dup			;
	movzx eax,byte [eax]	;
	inc dword [esi]		;
	call type		; Display the name of the word
	upsh 9			; Followed by a TAB
	call emit		;
	upsh 32			; And then a SPACE
	call emit		;
	jmp .a			;
.end:	pop ebx			;
	upsh 10			; At the end, put go to a new line
	call emit		;
next
; ----------------------------------------------------------------------
code 'reset', reset		; This word is used to reset the stack
	mov esi, s0		; to the proper starting point
next
code 'restart', restart		;
	mov dword [h], h0	;
	mov dword [d], d0 	;
	mov dword [base], 10	;
	upsh [save.flast]	;
	upop [flast]		;
	upsh [save.mlast]	;
	upop [mlast]		;
	upsh [save.last]	;
	upop [last]		;
	call reset		;
	call init_retroforth	;
next				;
code 'cmove', _cmove		;
	mov ecx,eax		;
        drop			;
	mov edi,eax		;
        drop			;
	push esi		;
	mov esi,eax		;
.a:     mov edx, [esi]		;
        mov [edi], edx		;
        inc edi			;
        inc esi			;
        loop .a			;
	pop esi			;
        drop			;
next
; ----------------------------------------------------------------------
mcode 'if', m_if		;
	upsh 0x74adc085		;
	call comma		;
	upsh [h]		;
	upop [if_resolve]	;
	upsh 0x0		;
	call comma1		;
next
; ----------------------------------------------------------------------
mcode 'then', m_then		;
	upsh [h]		;
	upsh [if_resolve]	;
	embed '- 1 -'		;
	upsh [if_resolve]	;
	upop edx		;
	mov [edx], al		;
	drop			;
next
; ----------------------------------------------------------------------
mcode 's"', s_quote		; Store a string into a word (sort of)
	upsh '"'		; * First, find a string
	call parse		; * Then move it to a safe area
	embed '2dup'		; * And finally compile the start+count
	upsh [stp]		;   into the word
	embed 'swap cmove swap drop'
	upsh [stp]		;
	call literal		; Put the address and count into the word
	add dword [stp], eax	; Advance SP by the count
	call literal		;
next
; ----------------------------------------------------------------------
var 'h', h, h0			; h (pointer to HERE)
var 'base', base, 10		; current numeric base
var 'sp', stp, h0+8000h		; String pointed, at end of code space
buffer		dd 0		; Buffer (for ports to use as needed)
source 		dd 0		; Evaluate from RAM or KBD
tin 		dd 0		; >IN
tp 		dd tib		; TP (pointer to input buffer)
d 		dd d0		; d (dictionary pointer)
if_resolve 	dd 0		; Used to resolve address for IF/THEN
bases		db 16,2,8,255	; $hex %bin &oct 'ascii
; ----------------------------------------------------------------------
save.flast: dd forth_link	;
save.mlast: dd macro_link	;
save.last:  dd flast		;
; ----------------------------------------------------------------------
flast dd forth_link		; Last word in 'forth'
mlast dd macro_link		; Last word in 'macro'
last dd flast			; Last word in dictionary
; ----------------------------------------------------------------------
retroforth_f: file "retroforth.f"	; Include the bootstrap code
retroforth_f.1:				;
; ----------------------------------------------------------------------
tib rb 1024				; Text Input Buffer
s0  rb 4*20				; Stack is 20 cells deep
d0  rb 2000h				; Dictionary space (was 10000h)
h0  rb 8000h				; Code space (was 80000h)
